import isEqual from "fast-deep-equal";
import {
  isPlayedWindup,
  memberIsWindup,
  playedMemberIsWindup,
  PlayedWindup,
  PlayedWindupMember,
  Windup,
  fastForward as vanillaFastForward,
  rewind as vanillaRewind,
} from "./Windup";

export type SuffixedWindupMember<ElementType, MetadataType> =
  | ElementType
  | SuffixedWindup<ElementType, MetadataType>;

export type SuffixedWindup<ElementType, MetadataType> = [
  Array<PlayedSuffixedWindupMember<ElementType, MetadataType>>,
  Array<SuffixedWindupMember<ElementType, MetadataType>>,
  Array<PlayedWindupMember<ElementType, MetadataType>>,
  MetadataType
];

export type PlayedSuffixedWindupMember<ElementType, MetadataType> =
  | ElementType
  | PlayedSuffixedWindup<ElementType, MetadataType>;

export type PlayedSuffixedWindup<ElementType, MetadataType> = [
  Array<PlayedSuffixedWindupMember<ElementType, MetadataType>>,
  Array<PlayedWindupMember<ElementType, MetadataType>>,
  MetadataType
];

export function isSuffixedWindup<ElementType, MetadataType>(
  windup:
    | SuffixedWindupMember<ElementType, MetadataType>
    | Windup<ElementType, MetadataType>
    | PlayedWindup<ElementType, MetadataType>
): windup is SuffixedWindup<ElementType, MetadataType> {
  // If it's not an array it can't be a windup
  if (!Array.isArray(windup)) {
    return false;
  }

  // If it has less or more than three members it's not a windup
  if (windup.length !== 4) {
    return false;
  }

  // If its first or second members are not arrays it's not a windup
  if (!Array.isArray(windup[0]) || !Array.isArray(windup[1])) {
    return false;
  }

  // Past here we just have to hope ElementType isn't a windup.
  return true;
}

export function isPlayedSuffixedWindup<ElementType, MetadataType>(
  windup:
    | SuffixedWindupMember<ElementType, MetadataType>
    | Windup<ElementType, MetadataType>
    | PlayedWindup<ElementType, MetadataType>
    | PlayedSuffixedWindupMember<ElementType, MetadataType>
): windup is PlayedSuffixedWindup<ElementType, MetadataType> {
  // If it's not an array it can't be a windup
  if (!Array.isArray(windup)) {
    return false;
  }

  // If it has less or more than three members it's not a windup
  if (windup.length !== 3) {
    return false;
  }

  // If its first or second members are not arrays it's not a windup
  if (!Array.isArray(windup[0]) || !Array.isArray(windup[1])) {
    return false;
  }

  // Past here we just have to hope ElementType isn't a windup.
  return true;
}

export function isUnplayed<ElementType, MetadataType>(
  windup: SuffixedWindup<ElementType, MetadataType>
): boolean {
  const [played, remaining] = windup;

  if (isPlayedSuffixedWindup(windup)) {
    return false;
  }

  if (played.length > 0) {
    return false;
  }

  return remaining.reduce((unplayed: boolean, member) => {
    if (isSuffixedWindup(member))
      if (isSuffixedWindup(windup) && unplayed) {
        return isUnplayed<ElementType, MetadataType>(member);
      }
    return unplayed;
  }, true);
}

export function playedSize<ElementType, MetadataType>(
  suffixedWindup:
    | SuffixedWindup<ElementType, MetadataType>
    | PlayedSuffixedWindup<ElementType, MetadataType>
): number {
  const [suffixedPlayed, remaining] = suffixedWindup;

  const sizeOfPlayed = suffixedPlayed.reduce((acc: number, member) => {
    if (isPlayedSuffixedWindup(member)) {
      return acc + playedSize(member);
    }

    return acc + 1;
  }, 0);

  const [firstRemaining] = remaining;

  if (firstRemaining && isSuffixedWindup(firstRemaining)) {
    const sizeFromFirstRemaining = playedSize(firstRemaining);

    return sizeOfPlayed + sizeFromFirstRemaining;
  }

  return sizeOfPlayed;
}

export function playedElementsWithSuffix<ElementType, MetadataType>(
  windup:
    | SuffixedWindup<ElementType, MetadataType>
    | PlayedSuffixedWindup<ElementType, MetadataType>
): Array<PlayedSuffixedWindupMember<ElementType, MetadataType>> {
  if (isPlayedSuffixedWindup(windup)) {
    const [played, suffix] = windup;

    return [
      ...played,
      ...suffix.map((member) => {
        if (playedMemberIsWindup(member)) {
          return playedWindupToSuffixed(member);
        }

        return member;
      }),
    ];
  }

  const [suffixedPlayed, remaining, suffix] = windup;

  let middle = remaining.reduce((acc, member) => {
    if (isSuffixedWindup(member)) {
      const [, , , metadata] = member;

      const asWindup = [
        [],
        playedElementsWithSuffix(member),
        metadata,
      ] as PlayedSuffixedWindup<ElementType, MetadataType>;

      return [...acc, asWindup];
    }

    return acc;
  }, [] as Array<PlayedSuffixedWindupMember<ElementType, MetadataType>>);

  return [
    ...suffixedPlayed,
    ...middle,
    ...suffix.map((member) => {
      if (playedMemberIsWindup(member)) {
        return playedWindupToSuffixed(member);
      }

      return member;
    }),
  ];
}

export function prev<ElementType, MetadataType>(
  suffixedWindup:
    | SuffixedWindup<ElementType, MetadataType>
    | PlayedSuffixedWindup<ElementType, MetadataType>
): SuffixedWindup<ElementType, MetadataType> {
  if (isPlayedSuffixedWindup(suffixedWindup)) {
    const [played, suffix, metadata] = suffixedWindup;

    const restPlayed = played.slice(0, -1);
    const lastPlayed = played[played.length - 1];

    if (!lastPlayed) {
      return [played, [], suffix, metadata];
    }

    if (isPlayedSuffixedWindup(lastPlayed)) {
      return [restPlayed, [prev(lastPlayed)], suffix, metadata];
    }

    return [restPlayed, [lastPlayed], suffix, metadata];
  }

  const [played, remaining, suffix, metadata] = suffixedWindup;

  const [firstRemaining, ...restRemaining] = remaining;

  if (!firstRemaining) {
    const restPlayed = played.slice(0, -1);
    const lastPlayed = played[played.length - 1];

    if (!lastPlayed) {
      return [played, [], suffix, metadata];
    }

    if (isPlayedSuffixedWindup(lastPlayed)) {
      return [restPlayed, [prev(lastPlayed)], suffix, metadata];
    }

    return [restPlayed, [lastPlayed], suffix, metadata];
  }

  // check if first remaining is an inprogress windup
  if (isSuffixedWindup(firstRemaining) && !isUnplayed(firstRemaining)) {
    return [played, [prev(firstRemaining), ...restRemaining], suffix, metadata];
  }

  // if it's not (element or unplayed)
  // take the last element of the played value
  const restPlayed = played.slice(0, -1);
  const lastPlayed = played[played.length - 1];

  if (isPlayedSuffixedWindup(lastPlayed)) {
    return [restPlayed, [prev(lastPlayed), ...remaining], suffix, metadata];
  }

  // if it's an element, just take it
  return [restPlayed, [lastPlayed, ...remaining], suffix, metadata];
}

export function next<ElementType, MetadataType>(
  suffixedWindup:
    | SuffixedWindup<ElementType, MetadataType>
    | PlayedSuffixedWindup<ElementType, MetadataType>
):
  | SuffixedWindup<ElementType, MetadataType>
  | PlayedSuffixedWindup<ElementType, MetadataType> {
  if (isPlayedSuffixedWindup(suffixedWindup)) {
    return suffixedWindup;
  }

  const [played, remaining, suffix, metadata] = suffixedWindup;
  const [firstRemaining, ...restRemaining] = remaining;

  if (!firstRemaining) {
    return [played, suffix, metadata];
  }

  if (isSuffixedWindup(firstRemaining)) {
    const nextFirstRemaining = next(firstRemaining);

    if (isPlayedSuffixedWindup(nextFirstRemaining)) {
      if (restRemaining.length === 0) {
        return [
          [...played, nextFirstRemaining],
          suffix,
          metadata,
        ] as PlayedSuffixedWindup<ElementType, MetadataType>;
      }

      return [
        [...played, nextFirstRemaining],
        restRemaining,
        suffix,
        metadata,
      ] as SuffixedWindup<ElementType, MetadataType>;
    }

    return [played, [nextFirstRemaining, ...restRemaining], suffix, metadata];
  }

  if (restRemaining.length === 0) {
    return [[...played, firstRemaining], suffix, metadata];
  }

  return [[...played, firstRemaining], restRemaining, suffix, metadata];
}

export function fastForward<E, M>(
  windup: SuffixedWindup<E, M> | PlayedSuffixedWindup<E, M>
): PlayedSuffixedWindup<E, M> {
  const forwardedWindup = next(windup);

  if (isPlayedSuffixedWindup(forwardedWindup)) {
    return forwardedWindup;
  }

  return fastForward(forwardedWindup);
}

export function windupToSuffixed<E, M>(
  windup: Windup<E, M>
): SuffixedWindup<E, M> {
  const [played, remainder, metadata] = windup;

  return [
    played.map((member) => {
      if (!playedMemberIsWindup(member)) {
        return member;
      }

      return playedWindupToSuffixed(member);
    }),
    remainder.map((member) => {
      if (!memberIsWindup(member)) {
        return member;
      }

      return windupToSuffixed(member);
    }),
    [],
    metadata,
  ];
}

export function playedWindupToSuffixed<E, M>(
  windup: PlayedWindup<E, M>
): PlayedSuffixedWindup<E, M> {
  const [played, metadata] = windup;

  return [
    played.map((member) => {
      if (!playedMemberIsWindup(member)) {
        return member;
      }

      return playedWindupToSuffixed(member);
    }),
    [],
    metadata,
  ];
}

export enum MemberKind {
  Played,
  InProgress,
}

// This serves two purposes:
// 1. Checking whether a member is a windup or not without using typeguards constantly
// 2. Letting us know whether standalone elements were played or in progress after pulling them out of their context.
export type AnnotatedMember<E, M> =
  | {
      member: E;
      kind: MemberKind.Played;
      isWindup: false;
    }
  | {
      member: E;
      kind: MemberKind.InProgress;
      isWindup: false;
    }
  | {
      member: Windup<E, M>;
      kind: MemberKind.InProgress;
      isWindup: true;
    }
  | {
      member: PlayedWindup<E, M>;
      kind: MemberKind.Played;
      isWindup: true;
    };

export function allMembers<E, M>(
  windup: Windup<E, M> | PlayedWindup<E, M>
): Array<AnnotatedMember<E, M>> {
  const played = isPlayedWindup(windup) ? windup[0] : windup[0];
  const remainder = !isPlayedWindup(windup) ? windup[1] : [];

  return [
    ...played.map((member) => {
      return playedMemberIsWindup(member)
        ? ({
            member,
            kind: MemberKind.Played,
            isWindup: true,
          } as {
            member: PlayedWindup<E, M>;
            kind: MemberKind.Played;
            isWindup: true;
          })
        : ({ member, kind: MemberKind.Played, isWindup: false } as {
            member: E;
            kind: MemberKind.Played;
            isWindup: false;
          });
    }),
    ...remainder.map((member) => {
      return memberIsWindup(member)
        ? ({
            member,
            kind: MemberKind.InProgress,
            isWindup: true,
          } as {
            member: Windup<E, M>;
            kind: MemberKind.InProgress;
            isWindup: true;
          })
        : ({ member, kind: MemberKind.InProgress, isWindup: false } as {
            member: E;
            kind: MemberKind.InProgress;
            isWindup: false;
          });
    }),
  ];
}

// Check if members are equivalent

function isMemberEquivalent<E, M>(
  memberA: AnnotatedMember<E, M>,
  memberB: AnnotatedMember<E, M>
): boolean {
  if (
    (!memberA.isWindup && memberB.isWindup) ||
    (memberA.isWindup && !memberB.isWindup)
  ) {
    return false;
  }

  if (memberA.isWindup && memberB.isWindup) {
    // TODO: look at metadata here
    const aMetadata = getMetadata(memberA.member);
    const bMetadata = getMetadata(memberB.member);

    if (!isEqual(aMetadata, bMetadata)) {
      return false;
    }

    const aMembers = allMembers(memberA.member);
    const bMembers = allMembers(memberB.member);

    if (aMembers.length !== bMembers.length) {
      return false;
    }

    let zipped = zipMembersFromLeft(aMembers, bMembers);

    return zipped.every(([a, b]) => (b ? isMemberEquivalent(a, b) : false));
  }

  return isEqual(memberA.member, memberB.member);
}

type ZipTuple<E, M> =
  | [AnnotatedMember<E, M>, AnnotatedMember<E, M>]
  | [AnnotatedMember<E, M>, undefined];

type Matches<E, M> = {
  matched: Array<AnnotatedMember<E, M>>;
  unmatched: Array<ZipTuple<E, M>>;
};

function matchesFromLeft<E, M>(zipped: ZipTuple<E, M>[]): Matches<E, M> {
  return zipped.reduce(
    (matches, [aMember, bMember]) => {
      if (
        bMember &&
        isMemberEquivalent(aMember, bMember) &&
        matches.unmatched.length === 0
      ) {
        return {
          ...matches,
          matched: [...matches.matched, aMember],
        };
      }

      return {
        ...matches,
        unmatched: [...matches.unmatched, [aMember, bMember] as ZipTuple<E, M>],
      };
    },
    { matched: [], unmatched: [] } as Matches<E, M>
  );
}

function matchesFromRight<E, M>(zipped: ZipTuple<E, M>[]): Matches<E, M> {
  return zipped.reduce(
    (matches, [aMember, bMember]) => {
      if (
        bMember &&
        isMemberEquivalent(aMember, bMember) &&
        matches.unmatched.length === 0
      ) {
        return {
          ...matches,
          matched: [aMember, ...matches.matched],
        };
      }

      return {
        ...matches,
        unmatched: [[aMember, bMember] as ZipTuple<E, M>, ...matches.unmatched],
      };
    },
    { matched: [], unmatched: [] } as Matches<E, M>
  );
}

function zipMembersFromLeft<E, M>(
  baseMembers: AnnotatedMember<E, M>[],
  againstMembers: AnnotatedMember<E, M>[]
): ZipTuple<E, M>[] {
  return baseMembers.map((member, i) => [member, againstMembers[i]]);
}

function zipMembersFromRight<E, M>(
  baseMembers: AnnotatedMember<E, M>[],
  againstMembers: AnnotatedMember<E, M>[]
): ZipTuple<E, M>[] {
  const zippedFromEnd: Array<ZipTuple<E, M>> = [];

  for (let i = 0; i <= baseMembers.length - 1; i++) {
    zippedFromEnd.push([
      baseMembers[baseMembers.length - 1 - i],
      againstMembers[againstMembers.length - 1 - i],
    ]);
  }

  return zippedFromEnd;
}

function prefixFromPair<E, M>(
  pair: [AnnotatedMember<E, M>, AnnotatedMember<E, M> | undefined]
): SuffixedWindup<E, M> | undefined {
  const [baseMember, againstMember] = pair;

  if (!baseMember.isWindup) {
    return undefined;
  }

  if (!againstMember?.isWindup) {
    return undefined;
  }

  const metadata = getMetadata(baseMember.member);

  const baseMembers = allMembers(baseMember.member);
  const againstMembers = allMembers(againstMember.member);
  const zipped = zipMembersFromLeft(baseMembers, againstMembers);

  const { matched, unmatched: unmatchedFromLeft } = matchesFromLeft(zipped);

  const played = matched.map((member) => {
    if (member.kind === MemberKind.InProgress && member.isWindup) {
      return playedWindupToSuffixed(vanillaFastForward(member.member));
    }

    if (member.kind === MemberKind.Played && member.isWindup) {
      return playedWindupToSuffixed(member.member);
    }

    return member.member;
  });

  const [firstUnmatchedFromLeft] = unmatchedFromLeft;
  const maybePrefix = prefixFromPair(firstUnmatchedFromLeft);

  const fromLeftBaseMembers = unmatchedFromLeft
    .map(([baseAnnotated]) => baseAnnotated)
    .map(annotatedMemberToRemaining);

  if (maybePrefix) {
    const remaining = fromLeftBaseMembers.slice(1, 0);

    return [played, [maybePrefix, ...remaining], [], metadata];
  }

  return [played, fromLeftBaseMembers, [], metadata];
}

function suffixFromPair<E, M>(
  pair: [AnnotatedMember<E, M>, AnnotatedMember<E, M> | undefined]
): SuffixedWindup<E, M> | undefined {
  const [baseMember, againstMember] = pair;

  if (!baseMember.isWindup) {
    return undefined;
  }

  if (!againstMember?.isWindup) {
    return undefined;
  }

  const metadata = getMetadata(baseMember.member);

  const baseMembers = allMembers(baseMember.member);
  const againstMembers = allMembers(againstMember.member);
  const zipped = zipMembersFromRight(baseMembers, againstMembers);

  const { matched, unmatched: unmatchedFromRight } = matchesFromRight(zipped);

  const lastUnmatchedFromRight =
    unmatchedFromRight[unmatchedFromRight.length - 1];
  const maybeSuffix = prefixFromPair(lastUnmatchedFromRight);

  const suffix = matched.map(annotatedMemberToPlayed);

  if (maybeSuffix) {
    const remaining = baseMembers
      .slice(0, unmatchedFromRight.length - 1)
      .map(annotatedMemberToRemaining);
    return [[], [...remaining, maybeSuffix], suffix, metadata];
  }

  const remaining = [...unmatchedFromRight]
    .map(([baseAnnotated]) => baseAnnotated)
    .map(annotatedMemberToRemaining);

  return [[], remaining, suffix, metadata];
}

function annotatedMemberToRemaining<E, M>(
  annotated: AnnotatedMember<E, M>
): SuffixedWindupMember<E, M> {
  if (annotated.isWindup && annotated.kind === MemberKind.Played) {
    // This shouldn't happen when this is used...
    return prev(playedWindupToSuffixed(annotated.member));
  }

  if (annotated.isWindup) {
    return windupToSuffixed(annotated.member);
  }

  return annotated.member;
}

function annotatedMemberToPlayed<E, M>(
  annotated: AnnotatedMember<E, M>
): PlayedWindupMember<E, M> {
  if (annotated.isWindup && annotated.kind === MemberKind.InProgress) {
    // This shouldn't happen when this is used...
    return vanillaFastForward(annotated.member);
  }

  return annotated.member;
}

function playedAndRemainingFromUnmatched<E, M>(
  fromLeft: ZipTuple<E, M>[],
  fromRight: ZipTuple<E, M>[]
): {
  played: PlayedSuffixedWindupMember<E, M>[];
  remaining: SuffixedWindupMember<E, M>[];
} {
  if (fromLeft.length === 0 || fromRight.length === 0) {
    return { played: [], remaining: [] };
  }

  const [[firstBaseFromLeft, firstAgainstFromLeft]] = fromLeft;
  const [firstBaseFromRight, firstAgainstFromRight] =
    fromRight[fromRight.length - 1];

  if (
    fromLeft.length === 1 &&
    fromRight.length === 1 &&
    firstBaseFromLeft.isWindup &&
    firstAgainstFromLeft?.isWindup
  ) {
    const merged = mergeWindupsToSuffixedWindup(
      firstBaseFromLeft.member,
      firstAgainstFromLeft.member
    );

    if (isPlayedSuffixedWindup(merged)) {
      return { played: [merged], remaining: [] };
    }

    return { played: [], remaining: [merged] };
  }

  const playedEls = fromLeft.filter(
    ([baseAnnotated]) => baseAnnotated.kind === MemberKind.Played
  );

  const played = playedEls
    .map(([baseAnnotated]) => baseAnnotated)
    .map((baseAnnotated) => {
      if (
        baseAnnotated.isWindup &&
        baseAnnotated.kind === MemberKind.InProgress
      ) {
        return fastForward(windupToSuffixed(baseAnnotated.member));
      }

      if (baseAnnotated.isWindup) {
        return playedWindupToSuffixed(baseAnnotated.member);
      }

      return baseAnnotated.member;
    });

  const shiftedLeft = fromLeft.slice(played.length);

  if (shiftedLeft.length === 0 && playedEls.length === 0) {
    return { played, remaining: [] };
  }

  if (shiftedLeft.length === 0) {
    const shiftedFromRight = fromRight.slice(0, playedEls.length);

    const [baseFromRight, againstFromRight] =
      shiftedFromRight[shiftedFromRight.length - 1];

    const maybeWindupSuffix = suffixFromPair([baseFromRight, againstFromRight]);

    if (maybeWindupSuffix) {
      const playedWithoutSuffix = played.slice(0, -1);

      return {
        played: [...playedWithoutSuffix, fastForward(maybeWindupSuffix)],
        remaining: [],
      };
    }

    return { played, remaining: [] };
  }

  const [[firstBaseFromShiftedLeft, firstAgainstFromShiftedLeft]] = shiftedLeft;

  const maybeWindupPrefix = prefixFromPair([
    firstBaseFromShiftedLeft,
    firstAgainstFromShiftedLeft,
  ]);
  const maybeWindupSuffix = suffixFromPair([
    firstBaseFromRight,
    firstAgainstFromRight,
  ]);

  const shiftedLeftBaseMembers = shiftedLeft
    .map(([baseAnnotated]) => baseAnnotated)
    .map(annotatedMemberToRemaining);

  if (maybeWindupPrefix && maybeWindupSuffix) {
    const remaining = shiftedLeftBaseMembers.slice(1, -1);

    return {
      played,
      remaining: [maybeWindupPrefix, ...remaining, maybeWindupSuffix],
    };
  }

  if (!maybeWindupPrefix && maybeWindupSuffix) {
    const remaining = shiftedLeftBaseMembers.slice(0, -1);

    return { played, remaining: [...remaining, maybeWindupSuffix] };
  }

  if (maybeWindupPrefix && !maybeWindupSuffix) {
    const remaining = shiftedLeftBaseMembers.slice(1);

    return { played, remaining: [maybeWindupPrefix, ...remaining] };
  }

  return { played, remaining: shiftedLeftBaseMembers };
}

export function mergeWindupsToSuffixedWindup<E, M>(
  baseWindup: Windup<E, M> | PlayedWindup<E, M>,
  againstWindup: Windup<E, M> | PlayedWindup<E, M>
): SuffixedWindup<E, M> | PlayedSuffixedWindup<E, M> {
  const metadata = getMetadata(baseWindup);

  // Get all the members and zip them up.
  const baseMembers = allMembers(baseWindup);
  const againstMembers = allMembers(againstWindup);
  const zipped = zipMembersFromLeft(baseMembers, againstMembers);

  const { matched, unmatched: unmatchedFromLeft } = matchesFromLeft(zipped);

  const [firstFromUnmatchedFromLeft] = unmatchedFromLeft;

  if (
    unmatchedFromLeft.length === 1 &&
    firstFromUnmatchedFromLeft[0].isWindup &&
    firstFromUnmatchedFromLeft[1]?.isWindup
  ) {
  }

  const played = matched.map((member) => {
    if (member.kind === MemberKind.InProgress && member.isWindup) {
      return playedWindupToSuffixed(vanillaFastForward(member.member));
    }

    if (member.kind === MemberKind.Played && member.isWindup) {
      return playedWindupToSuffixed(member.member);
    }

    return member.member;
  });

  const lengthToZipRight = baseMembers.length - played.length;

  const zippedRight = zipMembersFromRight(baseMembers, againstMembers).slice(
    0,
    lengthToZipRight
  );

  const { matched: matchedFromRight, unmatched: unmatchedZippedFromRight } =
    matchesFromRight(zippedRight);

  const unmatchedZippedFromLeft = unmatchedFromLeft.slice(
    0,
    unmatchedZippedFromRight.length
  );

  const { played: playedFromUnmatched, remaining } =
    playedAndRemainingFromUnmatched(
      unmatchedZippedFromLeft,
      unmatchedZippedFromRight
    );

  const suffix = matchedFromRight.map((member) => {
    if (member.kind === MemberKind.InProgress && member.isWindup) {
      return vanillaFastForward(member.member);
    }

    if (member.kind === MemberKind.Played && member.isWindup) {
      return member.member;
    }

    return member.member;
  });

  if (remaining.length === 0) {
    return [[...played, ...playedFromUnmatched], suffix, metadata];
  }

  return [[...played, ...playedFromUnmatched], remaining, suffix, metadata];
}

export function getMetadata<E, M>(
  windup: Windup<E, M> | PlayedWindup<E, M>
): M {
  return isPlayedWindup(windup) ? windup[1] : windup[2];
}

export function lastPlayedSuffixedMember<ElementType, MetadataType>(
  windup:
    | SuffixedWindup<ElementType, MetadataType>
    | PlayedSuffixedWindup<ElementType, MetadataType>
): PlayedSuffixedWindupMember<ElementType, MetadataType> | undefined {
  if (isPlayedSuffixedWindup(windup)) {
    const [played] = windup;
    const last = played[played.length - 1];
    return last;
  }

  const [played, remaining] = windup;

  const playedFromRemaining = remaining.reduce(
    (playedEl: ElementType | undefined, member) => {
      if (isSuffixedWindup(member)) {
        if (!isUnplayed(member)) {
          return lastPlayedSuffixedElement(member);
        }
      }
      return playedEl;
    },
    undefined
  );

  if (playedFromRemaining) {
    return playedFromRemaining;
  }

  const last = played[played.length - 1];
  return last;
}

export function lastPlayedSuffixedElement<ElementType, MetadataType>(
  windup:
    | SuffixedWindup<ElementType, MetadataType>
    | PlayedSuffixedWindup<ElementType, MetadataType>
): ElementType | undefined {
  const lastPlayed = lastPlayedSuffixedMember(windup);

  if (isPlayedSuffixedWindup(lastPlayed)) {
    return lastPlayedSuffixedElement(lastPlayed);
  }

  return lastPlayed;
}

export function nextSuffixedElement<ElementType, MetadataType>(
  windup:
    | SuffixedWindup<ElementType, MetadataType>
    | PlayedSuffixedWindup<ElementType, MetadataType>
): ElementType | undefined {
  if (isPlayedSuffixedWindup(windup)) {
    return undefined;
  }

  const [, remaining] = windup;

  const [nextVal] = remaining;

  if (isSuffixedWindup(nextVal)) {
    return nextSuffixedElement(nextVal);
  }

  return nextVal;
}

export function unsuffixWindup<E, M>(
  suffixed: SuffixedWindup<E, M>
): Windup<E, M> {
  const [played, remaining, suffix, metadata] = suffixed;

  const nextPlayed = played.map((member) => {
    if (isPlayedSuffixedWindup(member)) {
      return unsuffixPlayedWindup(member);
    }

    return member;
  });

  const nextRemaining = remaining.map((member) => {
    if (isSuffixedWindup(member)) {
      return unsuffixWindup(member);
    }

    return member;
  });

  const suffixAsMoreRemaining = suffix.map((member) => {
    if (playedMemberIsWindup(member)) {
      return vanillaRewind(member);
    }

    return member;
  });

  return [nextPlayed, [...nextRemaining, ...suffixAsMoreRemaining], metadata];
}

export function unsuffixPlayedWindup<E, M>(
  suffixed: PlayedSuffixedWindup<E, M>
): PlayedWindup<E, M> {
  const [played, suffix, metadata] = suffixed;

  const nextPlayed = played.map((member) => {
    if (isPlayedSuffixedWindup(member)) {
      return unsuffixPlayedWindup(member);
    }

    return member;
  });

  return [[...nextPlayed, ...suffix], metadata];
}
