// Possible member of a windup, e.g. a character or another windup.
export type WindupMember<ElementType, MetadataType> =
  | ElementType
  | Windup<ElementType, MetadataType>;

// A windup that has not yet been exhausted
export type Windup<ElementType, MetadataType> = [
  Array<PlayedWindupMember<ElementType, MetadataType>>,
  Array<WindupMember<ElementType, MetadataType>>,
  MetadataType
];

// Possible member of a windup, e.g. a character or another windup.
export type PlayedWindupMember<ElementType, MetadataType> =
  | ElementType
  | PlayedWindup<ElementType, MetadataType>;

// A windup that has exhausted all elements
export type PlayedWindup<ElementType, MetadataType> = [
  Array<PlayedWindupMember<ElementType, MetadataType>>,
  MetadataType
];

// Is the windup of type PlayedWindup?
export function isPlayedWindup<ElementType, MetadataType>(
  windup:
    | PlayedWindup<ElementType, MetadataType>
    | Windup<ElementType, MetadataType>
): windup is PlayedWindup<ElementType, MetadataType> {
  // If it's not an array it can't be a windup
  if (!Array.isArray(windup)) {
    return false;
  }

  // If it has less or more than three members it's not a windup
  if (windup.length !== 2) {
    return false;
  }
  
  // If its first member is not  anarrays it's not a windup
  if (!Array.isArray(windup[0])) {
    return false;
  }

  return true;
}

// Is the member a windup, or just a member type?
export function memberIsWindup<ElementType, MetadataType>(
  member: WindupMember<ElementType, MetadataType>
): member is Windup<ElementType, MetadataType> {
  // If it's not an array it can't be a windup
  if (!Array.isArray(member)) {
    return false;
  }

  // If it has less or more than three members it's not a windup
  if (member.length !== 3) {
    return false;
  }

  // If its first or second members are not arrays it's not a windup
  if (!Array.isArray(member[0]) || !Array.isArray(member[1])) {
    return false;
  }

  // Past here we just have to hope ElementType isn't a windup.
  return true;
}

// Is the member a windup, or just a member type?
export function playedMemberIsWindup<ElementType, MetadataType>(
  member: PlayedWindupMember<ElementType, MetadataType>
): member is PlayedWindup<ElementType, MetadataType> {
  // If it's not an array it can't be a windup
  if (!Array.isArray(member)) {
    return false;
  }

  // If it has less or more than three members it's not a windup
  if (member.length !== 2) {
    return false;
  }

  // If its first member is not an array it's not a windup
  if (!Array.isArray(member[0])) {
    return false;
  }

  // Past here we just have to hope ElementType isn't a windup.
  return true;
}

// Create a windup from a given string
// "Hello", metadata -> [[], ["H", "e", "l", "l", "o"], metadata]
export function windupFromString<MetadataType>(
  str: string,
  metadata: MetadataType
): Windup<string, MetadataType> {
  return [[], str.split(""), metadata];
}

export function newWindup<ElementType, MetadataType>(
  arg: Array<WindupMember<ElementType, MetadataType>>,
  metadata: MetadataType
): Windup<ElementType, MetadataType> {
  return [[], arg, metadata];
}

// Checks if the windup is unplayed
// INCLUDING any nested windups in the remaining part of the windup
export function isUnplayed<ElementType, MetadataType>(
  windup:
    | Windup<ElementType, MetadataType>
    | PlayedWindup<ElementType, MetadataType>
): boolean {
  if (isPlayedWindup(windup)) {
    return false;
  }

  const [played, remaining] = windup;

  if (played.length > 0) {
    return false;
  }

  return remaining.reduce((unplayed: boolean, member) => {
    if (memberIsWindup(member))
      if (memberIsWindup(windup) && unplayed) {
        return isUnplayed<ElementType, MetadataType>(member);
      }
    return unplayed;
  }, true);
}

// Checks if a windup has any remaining parts
export function isFinished<ElementType, MetadataType>(
  windup:
    | Windup<ElementType, MetadataType>
    | PlayedWindup<ElementType, MetadataType>
): boolean {
  if (isPlayedWindup(windup)) {
    return true;
  }

  const [, remaining] = windup;

  return remaining.length === 0;
}

// Turn a Windup into a PlayedWindup
export function fastForward<ElementType, MetadataType>(
  windup: Windup<ElementType, MetadataType> | PlayedWindup<ElementType, MetadataType>
): PlayedWindup<ElementType, MetadataType> {
  if (isPlayedWindup(windup)) {
    return windup
  }
  
  const forwardedWindup = next(windup);
  if (isPlayedWindup(forwardedWindup)) {
    return forwardedWindup;
  }
  return fastForward(forwardedWindup);
}

// Turn an in-progress Windup or PlayedWindup into an unplayed Windup
export function rewind<ElementType, MetadataType>(
  windup:
    | Windup<ElementType, MetadataType>
    | PlayedWindup<ElementType, MetadataType>
): Windup<ElementType, MetadataType> {
  if (!isPlayedWindup(windup) && isUnplayed(windup)) {
    return windup;
  }

  const played = windup[0];
  const metadata = isPlayedWindup(windup) ? windup[1] : windup[2];

  const mapRewind = (
    member: WindupMember<ElementType, MetadataType>
  ): WindupMember<ElementType, MetadataType> => {
    if (memberIsWindup(member)) {
      return rewind(member);
    }

    return member;
  };

  const playedMapRewind = (
    member: PlayedWindupMember<ElementType, MetadataType>
  ) => {
    if (playedMemberIsWindup(member)) {
      return rewind(member);
    }

    return member;
  };

  if (isPlayedWindup(windup)) {
    return [[], played.map(playedMapRewind), metadata];
  }

  const remaining = windup[1];

  return [
    [],
    [...played.map(playedMapRewind), ...remaining.map(mapRewind)],
    metadata,
  ];
}

export function windupAsString<MetadataType>(
  windup: Windup<string, MetadataType> | PlayedWindup<string, MetadataType>
): string {
  const played = windup[0];
  const remaining = isPlayedWindup(windup) ? [] : windup[1];

  type MemberType = WindupMember<string, MetadataType>;
  const stringify = (member: MemberType): MemberType => {
    if (memberIsWindup(member)) {
      return windupAsString(member);
    }
    return member;
  };

  type PlayedMemberType = PlayedWindupMember<string, MetadataType>;
  const stringifyPlayed = (member: PlayedMemberType): PlayedMemberType => {
    if (playedMemberIsWindup(member)) {
      return windupAsString(member);
    }
    return member;
  };
  return [
    played.map(stringifyPlayed).join(""),
    remaining.map(stringify).join(""),
  ].join("");
}

export function lastPlayedMember<ElementType, MetadataType>(
  windup:
    | Windup<ElementType, MetadataType>
    | PlayedWindup<ElementType, MetadataType>
): PlayedWindupMember<ElementType, MetadataType> | undefined {
  if (isPlayedWindup(windup)) {
    const [played] = windup;
    const last = played[played.length - 1];
    return last;
  }

  const [played, remaining] = windup;

  const playedFromRemaining = remaining.reduce(
    (playedEl: ElementType | undefined, member) => {
      if (memberIsWindup(member)) {
        if (!isUnplayed(member)) {
          return lastPlayedElement(member);
        }
      }
      return playedEl;
    },
    undefined
  );

  if (playedFromRemaining) {
    return playedFromRemaining;
  }

  const last = played[played.length - 1];
  return last;
}

export function lastPlayedElement<ElementType, MetadataType>(
  windup:
    | Windup<ElementType, MetadataType>
    | PlayedWindup<ElementType, MetadataType>
): ElementType | undefined {
  const lastPlayed = lastPlayedMember(windup);

  if (playedMemberIsWindup(lastPlayed)) {
    return lastPlayedElement(lastPlayed);
  }

  return lastPlayed;
}

// Return the played elements of a windup
// INCLUDING any played elements from nested windups in the remaining part of the windup
export function playedElements<ElementType, MetadataType>(
  windup:
    | Windup<ElementType, MetadataType>
    | PlayedWindup<ElementType, MetadataType>
): Array<PlayedWindupMember<ElementType, MetadataType>> {
  if (isPlayedWindup(windup)) {
    const [played] = windup;

    return played;
  }

  const [played, remaining] = windup;

  const [firstRemaining] = remaining;

  if (memberIsWindup(firstRemaining) && !isUnplayed(firstRemaining)) {
    const [, , firstRemainingMetadata] = firstRemaining;

    return [
      ...played,
      [playedElements(firstRemaining), firstRemainingMetadata],
    ] as Array<PlayedWindupMember<ElementType, MetadataType>>;
  }

  return played;
}

export function nextElement<ElementType, MetadataType>(
  windup:
    | Windup<ElementType, MetadataType>
    | PlayedWindup<ElementType, MetadataType>
): ElementType | undefined {
  if (isPlayedWindup(windup)) {
    return undefined;
  }

  const [, remaining] = windup;

  const [nextVal] = remaining;

  if (memberIsWindup(nextVal)) {
    return nextElement(nextVal);
  }

  return nextVal;
}

export function prev<ElementType, MetadataType>(
  windup:
    | Windup<ElementType, MetadataType>
    | PlayedWindup<ElementType, MetadataType>
): Windup<ElementType, MetadataType> {
  const [played] = windup;

  const metadata = isPlayedWindup(windup) ? windup[1] : windup[2];

  const remaining = isPlayedWindup(windup) ? [] : windup[1];

  if (!isPlayedWindup(windup)) {
    const [firstRemaining, ...restRemaining] = remaining;

    if (memberIsWindup(firstRemaining)) {
      return [played, [prev(firstRemaining), ...restRemaining], metadata];
    }
  }

  // if it's not (element or unplayed)
  // take the last element of the played value
  const restPlayed = played.slice(0, -2);
  const lastPlayed = played[played.length - 1];

  if (playedMemberIsWindup(lastPlayed)) {
    return [restPlayed, [prev(lastPlayed), ...remaining], metadata];
  }

  return [restPlayed, [lastPlayed, ...remaining], metadata];
}

export function next<ElementType, MetadataType>(
  windup: Windup<ElementType, MetadataType>
): Windup<ElementType, MetadataType> | PlayedWindup<ElementType, MetadataType> {
  // start
  // [[], [[[], ["h", "i"]], [[], ["n", "o"]]]]

  // next
  // [[], [[["h"], ["i"]], ["", ["n", "o"]]]]

  // next
  // [["h", "i"], [["n", "o"]]]

  // next
  // [["h", "i"] ["n", "o"]], []]

  if (isPlayedWindup(windup)) {
    return windup;
  }

  const [played, remaining, metadata] = windup;
  const [firstRemaining, ...restRemaining] = remaining;
  
  if (!firstRemaining) {
    return [played, metadata]
  }

  if (memberIsWindup(firstRemaining)) {
    const nextFirstRemaining = next(firstRemaining);

    if (isPlayedWindup(nextFirstRemaining)) {
      if (restRemaining.length === 0) {
        return [[...played, nextFirstRemaining], metadata] as PlayedWindup<
          ElementType,
          MetadataType
        >;
      }

      return [
        [...played, nextFirstRemaining],
        restRemaining,
        metadata,
      ] as Windup<ElementType, MetadataType>;
    }

    return [played, [nextFirstRemaining, ...restRemaining], metadata];
  }

  if (restRemaining.length === 0) {
    return [[...played, firstRemaining], metadata];
  }

  return [[...played, firstRemaining], restRemaining, metadata];
}
