import * as React from "react";
import {
  isPlayedSuffixedWindup,
  isSuffixedWindup,
  PlayedSuffixedWindup,
  SuffixedWindup,
} from "../SuffixedWindup";
import {
  isPlayedWindup,
  memberIsWindup,
  PlayedWindup,
  Windup,
} from "../Windup";
import { SuffixedHookMetadata } from "./useRewindup";
import { HookMetadata } from "./useWindup";

type MsProp = {
  ms: number;
  reverseMs?: number;
};

type ReverseMsProp = {
  ms?: number;
  reverseMs: number;
};

type GetPaceProp = {
  getPace: (char: string) => number;
  getReversePace?: (char: string) => number;
};

type ReverseGetPaceProp = {
  getPace?: (char: string) => number;
  getReversePace: (char: string) => number;
};

type PaceProps = MsProp | ReverseMsProp | GetPaceProp | ReverseGetPaceProp;

export function defaultGetPace(
  lastChar: string,
  nextChar: string | undefined
): number {
  switch (lastChar) {
    case "—":
    case "…":
      return 200;
    case ".":
    case ",":
      return 150;
    case "?":
    case "!":
      if (nextChar !== "!" && nextChar !== "?") {
        return 150;
      }
      return 20;
    case "-":
    case " ":
    case "\n":
      return 0;
    default:
      return 20;
  }
}

const Pace: React.FC<PaceProps> = ({ children }) => {
  return <>{children}</>;
};

export function isPaceElement(
  element: React.ReactElement
): element is React.ReactElement<PaceProps> {
  return element.type === Pace;
}

export function isMsProp(props: PaceProps): props is MsProp {
  if ("ms" in props) {
    return true;
  }
  return false;
}

export function isReverseMsProp(props: PaceProps): props is ReverseMsProp {
  if ("reverseMs" in props) {
    return true;
  }

  return false;
}

export function isReverseGetPaceProp(
  props: PaceProps
): props is ReverseGetPaceProp {
  if ("reverseGetPace" in props) {
    return true;
  }

  return false;
}

export function paceMetadataFromProps(props: PaceProps):
  | {
      pace: (char: string) => number;
      reversePace?: (char: string) => number;
    }
  | {
      pace?: (char: string) => number;
      reversePace: (char: string) => number;
    } {
  if (isMsProp(props)) {
    const { ms, reverseMs } = props;

    return {
      pace: () => ms,
      ...(reverseMs ? { reversePace: () => reverseMs } : {}),
    };
  }

  if (isReverseMsProp(props)) {
    const { ms, reverseMs } = props;

    return {
      reversePace: () => reverseMs,
      ...(ms ? { pace: () => ms } : {}),
    };
  }

  if (isReverseGetPaceProp(props)) {
    return {
      reversePace: props.getReversePace,
      ...(props.getPace ? { pace: props.getPace } : {}),
    };
  }

  return {
    pace: props.getPace,
    ...(props.getReversePace ? { reversePace: props.getReversePace } : {}),
  };
}

export function paceFromWindup<MetadataType extends HookMetadata>(
  windup: Windup<string, MetadataType> | PlayedWindup<string, MetadataType>,
  parentPace?: (char: string, nextChar: string | undefined) => number
): ((char: string, nextChar: string | undefined) => number) | undefined {
  if (isPlayedWindup(windup)) {
    return undefined;
  }

  const [, remaining, metadata] = windup;

  const [firstRemaining] = remaining;

  if (firstRemaining && memberIsWindup(firstRemaining)) {
    return paceFromWindup(firstRemaining, metadata.pace || parentPace);
  }

  return metadata.pace || parentPace;
}

export function paceFromSuffixedWindup<
  MetadataType extends SuffixedHookMetadata
>(
  windup:
    | SuffixedWindup<string, MetadataType>
    | PlayedSuffixedWindup<string, MetadataType>,
  rewinding: boolean,
  parentPace?: (char: string, nextChar: string | undefined) => number
): ((char: string, nextChar: string | undefined) => number) | undefined {
  if (isPlayedSuffixedWindup(windup)) {
    return undefined;
  }

  const [, remaining, , metadata] = windup;

  const [firstRemaining] = remaining;

  const paceToUse = rewinding ? metadata.reversePace : metadata.pace;

  if (firstRemaining && isSuffixedWindup(firstRemaining)) {
    return paceFromSuffixedWindup(
      firstRemaining,
      rewinding,
      paceToUse || parentPace
    );
  }

  return paceToUse || parentPace;
}

export default Pace;
