import * as React from "react";
import { isPlayedWindup, newWindup, PlayedWindup, Windup } from "../Windup";
import {
  isPlayedSuffixedWindup,
  lastPlayedSuffixedElement,
  mergeWindupsToSuffixedWindup,
  next,
  nextSuffixedElement,
  PlayedSuffixedWindup,
  playedWindupToSuffixed,
  prev,
  SuffixedWindup,
  unsuffixPlayedWindup,
  unsuffixWindup,
  windupToSuffixed,
  fastForward,
  playedSize,
} from "../SuffixedWindup";
import { HookMetadata } from "./useWindup";
import { defaultGetPace, paceFromSuffixedWindup } from "./Pace";
import { onCharsFromSuffixedWindup } from "./OnChar";

function usePrevious<Type>(value: Type): Type | undefined {
  const ref = React.useRef<Type>();

  React.useEffect(() => {
    ref.current = value;
  }, [value]);

  return ref.current;
}

export interface SuffixedHookMetadata extends HookMetadata {
  reversePace?: (char: string, prevChar: string | undefined) => number;
}

type InitState<MetadataType> = {
  toPlay:
    | SuffixedWindup<string, MetadataType>
    | PlayedSuffixedWindup<string, MetadataType>;
};

type RewindupState<MetadataType> = {
  toRewind:
    | SuffixedWindup<string, MetadataType>
    | PlayedSuffixedWindup<string, MetadataType>;
  toResume:
    | SuffixedWindup<string, MetadataType>
    | PlayedSuffixedWindup<string, MetadataType>;
};

type RewindupReducerState<MetadataType> =
  | InitState<MetadataType>
  | RewindupState<MetadataType>;

type ReducerType<M> = (
  prevState: RewindupReducerState<M>,
  action: RewindupReducerAction<M>
) => RewindupReducerState<M>;

function isRewindupState<M>(
  state: RewindupReducerState<M>
): state is RewindupState<M> {
  if ("toRewind" in state) {
    return true;
  }

  return false;
}

function isRewinding<M>(state: RewindupReducerState<M>): boolean {
  return isRewindupState(state)
    ? playedSize(state.toRewind) > playedSize(state.toResume)
    : false;
}

function getActiveWindup<M>(
  state: RewindupReducerState<M>
): SuffixedWindup<string, M> | PlayedSuffixedWindup<string, M> {
  return isRewindupState(state)
    ? isRewinding(state)
      ? state.toRewind
      : state.toResume
    : state.toPlay;
}

function getNextBaseWindup<M>(
  state: RewindupReducerState<M>
): Windup<string, M> | PlayedWindup<string, M> {
  if (!isRewindupState(state)) {
    return isPlayedSuffixedWindup(state.toPlay)
      ? unsuffixPlayedWindup(state.toPlay)
      : unsuffixWindup(state.toPlay);
  }

  if (isRewinding(state)) {
    return isPlayedSuffixedWindup(state.toRewind)
      ? unsuffixPlayedWindup(state.toRewind)
      : unsuffixWindup(state.toRewind);
  }

  return isPlayedSuffixedWindup(state.toResume)
    ? unsuffixPlayedWindup(state.toResume)
    : unsuffixWindup(state.toResume);
}

type RewindupReducerAction<MetadataType> =
  | {
      type: "start-rewindup";
      baseWindup:
        | Windup<string, MetadataType>
        | PlayedWindup<string, MetadataType>;
      againstWindup:
        | Windup<string, MetadataType>
        | PlayedWindup<string, MetadataType>;
    }
  | {
      type: "proceed";
    }
  | { type: "fast-forward" };

function rewindupReducer<M>(
  state: RewindupReducerState<M>,
  action: RewindupReducerAction<M>
): RewindupReducerState<M> {
  switch (action.type) {
    case "start-rewindup": {
      const toRewind = mergeWindupsToSuffixedWindup(
        action.baseWindup,
        action.againstWindup
      );
      const toResume = mergeWindupsToSuffixedWindup(
        action.againstWindup,
        action.baseWindup
      );

      return {
        toRewind,

        toResume,
      };
    }

    case "proceed": {
      if (!isRewindupState(state)) {
        if (isPlayedSuffixedWindup(state.toPlay)) {
          return state;
        }

        return {
          toPlay: next(state.toPlay),
        };
      }

      if (isRewinding(state)) {
        const nextToRewind = prev(state.toRewind);

        return {
          ...state,
          toRewind: nextToRewind,
        };
      }

      const nextToResume = next(state.toResume);

      return {
        ...state,
        toResume: nextToResume,
      };
    }
    case "fast-forward": {
      return { toPlay: fastForward(getActiveWindup(state)) };
    }

    default:
      return state;
  }
}

export default function useRewindup<MetadataType extends SuffixedHookMetadata>(
  windupInit: Windup<string, MetadataType> | PlayedWindup<string, MetadataType>,
  options: {
    onFinished?: () => void;
    onRewindStart?: () => void;
  }
): {
  windup:
    | SuffixedWindup<string, MetadataType>
    | PlayedSuffixedWindup<string, MetadataType>;
  skip: () => void;
  rewind: () => void;
  isFinished: boolean;
} {
  const [state, dispatch] = React.useReducer<ReducerType<MetadataType>>(
    rewindupReducer,
    {
      toPlay: isPlayedWindup(windupInit)
        ? playedWindupToSuffixed(windupInit)
        : windupToSuffixed(windupInit),
    }
  );

  // When the windup changes, onChar should fire
  React.useEffect(() => {
    if (!state) {
      return;
    }

    const windupToGetOnCharFrom = getActiveWindup(state);

    const onChars = onCharsFromSuffixedWindup(windupToGetOnCharFrom);
    const lastEl = lastPlayedSuffixedElement(windupToGetOnCharFrom);

    if (onChars.length > 0 && lastEl) {
      onChars.forEach((onChar) => {
        onChar(lastEl, isRewindupState(state));
      });
    }
  }, [state]);

  const timeoutRef = React.useRef<NodeJS.Timeout | null>(null);

  // If windup finishes, the onFinished should fire
  const windupIsFinished = isPlayedSuffixedWindup(getActiveWindup(state));
  const didFinishOnce = React.useRef(false);

  React.useEffect(() => {
    // Put this in a new context so that the windup finishes visually before firing this
    if (windupIsFinished && !didFinishOnce.current) {
      const timeout = setTimeout(() => {
        if (options.onFinished) {
          options.onFinished();
        }
        didFinishOnce.current = true;
      }, 0);
      return () => {
        clearTimeout(timeout);
      };
    }
  }, [windupIsFinished, options.onFinished]);

  // Proceed the windup
  React.useEffect(() => {
    const isPlayed = isPlayedSuffixedWindup(
      isRewindupState(state) ? state.toResume : state.toPlay
    );
    const isRewindDone = isRewindupState(state)
      ? playedSize(state.toResume) >= playedSize(state.toRewind)
      : true;

    if (!state || (isPlayed && isRewindDone)) {
      return;
    }

    const windupToGetPaceFrom = getActiveWindup(state);

    const getPace =
      paceFromSuffixedWindup(windupToGetPaceFrom, isRewinding(state)) ||
      defaultGetPace;

    const lastEl = lastPlayedSuffixedElement(windupToGetPaceFrom);
    const nextEl = nextSuffixedElement(windupToGetPaceFrom);

    const timeout = lastEl ? getPace(lastEl, nextEl) : 0;

    timeoutRef.current = setTimeout(() => {
      dispatch({ type: "proceed" });
    }, timeout);
    return () => {
      if (timeoutRef.current) {
        clearTimeout(timeoutRef.current);
      }
    };
  }, [state]);

  const prevWindupInit = usePrevious(windupInit);

  React.useEffect(() => {
    didFinishOnce.current = false
    
    if (prevWindupInit === undefined) {
      return;
    }

    if (prevWindupInit === windupInit) {
      return;
    }

    if (options.onRewindStart) {
      options.onRewindStart();
    }

    dispatch({
      type: "start-rewindup",
      baseWindup: getNextBaseWindup(state),
      againstWindup: windupInit,
    });
  }, [windupInit, prevWindupInit]);

  const skip = React.useCallback(() => {
    dispatch({ type: "fast-forward" });
  }, []);

  const rewind = React.useCallback(() => {
    dispatch({
      type: "start-rewindup",
      baseWindup: newWindup<string, MetadataType>([], {} as MetadataType),
      againstWindup: windupInit,
    });
    didFinishOnce.current = false
  }, []);

  return {
    windup: getActiveWindup(state),
    skip,
    rewind,
    isFinished: windupIsFinished,
  };
}
